use crate::smallint::SmallUintType;
use crate::SmallUint;
use core::mem::ManuallyDrop;
use core::ops::{BitAnd, BitOr, BitXor};
use core::ops::{BitAndAssign, BitOrAssign, BitXorAssign};

macro_rules! logic_op {
    (
        $imp:ident, $imp_assign:ident, $typ:ident, $typ_inner:ident, $fun:ident, $fun_assign:ident;
        $i:ident, $j:ident, $r:ident, $s:ident;
        $inline_heap:tt, $heap_heap:tt
    ) => {
        fn $fun(a: &$typ, b: &$typ) -> $typ {
            match (&a.0, &b.0) {
                (&$typ_inner::Inline($i), &$typ_inner::Inline($j)) => {
                    $typ($typ_inner::Inline($i.$fun($j)))
                }

                (&$typ_inner::Inline($i), &$typ_inner::Heap(($r, $s)))
                | (&$typ_inner::Heap(($r, $s)), &$typ_inner::Inline($i)) => {
                    $inline_heap
                }

                (&$typ_inner::Heap(($r, $s)), &$typ_inner::Heap(($i, $j))) => {
                    $heap_heap
                }
            }
        }

        impl<'a, 'b> $imp<&'a $typ> for &'b $typ {
            type Output = $typ;

            fn $fun(self, rhs: &$typ) -> Self::Output {
                $fun(self, rhs)
            }
        }

        impl<'a> $imp<$typ> for &'a $typ {
            type Output = $typ;

            fn $fun(self, rhs: $typ) -> Self::Output {
                self.$fun(&rhs)
            }
        }

        impl<'a> $imp<&'a $typ> for $typ {
            type Output = $typ;

            fn $fun(self, rhs: &$typ) -> Self::Output {
                (&self).$fun(rhs)
            }
        }

        impl $imp<$typ> for $typ {
            type Output = $typ;

            fn $fun(self, rhs: $typ) -> Self::Output {
                (&self).$fun(&rhs)
            }
        }

        impl<'a> $imp_assign<&'a SmallUint> for $typ {
            fn $fun_assign(&mut self, rhs: &'a $typ) {
                *self = (&*self).$fun(rhs);
            }
        }

        impl $imp_assign<$typ> for $typ {
            fn $fun_assign(&mut self, rhs: $typ) {
                *self = (&*self).$fun(rhs);
            }
        }
    };
    (
        $imp:ident, $imp_assign:ident, $typ:ident, $typ_inner:ident, $fun:ident, $fun_assign:ident;
        $i:ident, $j:ident, $r:ident, $s:ident, $slice:ident, $slice1:ident, $slice2:ident, $min:ident, $res:ident;
        $inline_heap_inner:tt, $heap_heap_create_res:tt, $heap_heap_return:tt
    ) => {
        logic_op! {
            $imp, $imp_assign, $typ, $typ_inner, $fun, $fun_assign;
            $i, $j, $r, $s;
            {
                let $slice = unsafe { core::slice::from_raw_parts($r, $s) };

                $inline_heap_inner
            },
            {
                let $slice1 = unsafe { core::slice::from_raw_parts($r, $s) };
                let $slice2 = unsafe { core::slice::from_raw_parts($i, $j) };

                let $min = std::cmp::min($slice1.len(), $slice2.len());

                #[allow(unused_mut)]
                let mut $res = $heap_heap_create_res;

                $heap_heap_return
            }
        }
    };
}


macro_rules! inline_heap_to_inline {
    ($fun:tt; $i:ident, $slice:ident) => {
        let mut j = 0u128;
        for i in 0..4 {
            j <<= 32;
            j |= $slice[3 - i] as u128;
        }
        SmallUint(SmallUintType::Inline($i.$fun(j)))
    }
}

macro_rules! inline_heap_to_heap {
    ($fun_assign:ident; $i:ident, $slice:ident) => {
        let mut retvec = $slice.to_vec();

        let mut v = $i;
        #[allow(clippy::needless_range_loop)]
        for r in 0..4 {
            retvec[r].$fun_assign(v as u32);

            v >>= 32;
        }

        let mut retslice = ManuallyDrop::new(retvec.into_boxed_slice());

        SmallUint(SmallUintType::Heap((retslice.as_mut_ptr(), retslice.len())))
    }
}


macro_rules! heap_heap_create_res_shortest {
    ($fun:ident; $min:ident, $slice1:ident, $slice2:ident) => {
        let mut res = Vec::with_capacity($min);
        for l in 0..$min {
            res.push($slice1[l].$fun($slice2[l]));
        }
        res
    }
}

macro_rules! heap_heap_create_res_longest {
    ($fun:ident; $slice1:ident, $slice2:ident, $min:ident) => {
        let mut res = if $slice1.len() > $slice2.len() {
            $slice1.to_vec()
        } else {
            $slice2.to_vec()
        };
        for l in 0..$min {
            res[l] = $slice1[l].$fun($slice2[l]);
        }
        res
    }
}


macro_rules! heap_heap_return_heap {
    ($res:ident) => {
        let mut slice = ManuallyDrop::new($res.into_boxed_slice());
        SmallUint(SmallUintType::Heap((slice.as_mut_ptr(), slice.len())))
    }
}

macro_rules! heap_heap_return_any {
    ($res:ident) => {
        while $res.len() != 1 && $res[$res.len() - 1] == 0 {
            $res.pop();
        }

        if $res.len() <= 4 {
            let mut r = 0u128;
            for t in 0..$res.len() {
                r <<= 32;
                r |= $res[$res.len() - 1 - t] as u128;
            }
            SmallUint(SmallUintType::Inline(r))
        } else {
            let mut slice = ManuallyDrop::new($res.into_boxed_slice());
            SmallUint(SmallUintType::Heap((slice.as_mut_ptr(), slice.len())))
        }
    }
}

logic_op! {
    BitAnd, BitAndAssign, SmallUint, SmallUintType, bitand, bitand_assign;
    i, j, r, s, slice, slice1, slice2, min, res;
    { inline_heap_to_inline! { bitand; i, slice } },
    { heap_heap_create_res_shortest! { bitand; min, slice1, slice2 } },
    { heap_heap_return_any! { res } }
}

logic_op! {
    BitOr, BitOrAssign, SmallUint, SmallUintType, bitor, bitor_assign;
    i, j, r, s, slice, slice1, slice2, min, res;
    { inline_heap_to_heap! { bitor_assign; i, slice } },
    { heap_heap_create_res_longest! { bitor; slice1, slice2, min }},
    { heap_heap_return_heap! { res } }
}

logic_op! {
    BitXor, BitXorAssign, SmallUint, SmallUintType, bitxor, bitxor_assign;
    i, j, r, s, slice, slice1, slice2, min, res;
    { inline_heap_to_heap! { bitxor_assign; i, slice } },
    { heap_heap_create_res_longest! { bitxor; slice1, slice2, min }},
    { heap_heap_return_any! { res } }
}


